from .module import Module


class Configuration(Module):
    dependencies = ['region']

    def __init__(self, **kwargs):
        super().__init__('configuration', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='configuration help')

        p = subp.add_parser('bucket', help='set FUKU bucket')
        p.add_argument('name', metavar='NAME', help='bucket name')
        p.set_defaults(configuration_handler=self.handle_bucket)

        p = subp.add_parser('ls', help='list configuration')
        p.set_defaults(configuration_handler=self.handle_list)

    def handle_bucket(self, args):
        self.bucket(args.name)

    def bucket(self, name):
        s3 = self.get_boto_resource('s3')
        bucket = s3.Bucket(name)
        try:
            bucket.load()
        except:
            bucket.create()
        self.store_set('bucket', name)

    def handle_list(self, args):
        self.list()

    def list(self):
        if 'bucket' in self.store:
            print(f'bucket: {self.store["bucket"]}')

    def get_my_context(self):
        ctx = {}
        if 'bucket' not in self.store:
            self.error('bucket not set')
        ctx['bucket'] = self.store['bucket']
        return ctx

    def save(self):
        cache = super().save()
        if 'bucket' in self.store:
            cache.update({
                'bucket': self.store['bucket']
            })
        return cache

    def load(self, cache):
        super().load(cache)
        if 'bucket' in cache:
            self.store['bucket'] = cache['bucket']
        else:
            try:
                del self.store['bucket']
            except KeyError:
                pass



import argparse

from .db import get_default_db, save_db


class Client(object):
    def __init__(self):
        self.modules = []
        self.parser = argparse.ArgumentParser()
        self.db = get_default_db()

    def add_module(self, module):
        module = module(db=self.db, client=self)
        for mod in self.modules:
            if mod.name == module.name:
                raise TypeError('duplicate modules: {}'.format(mod.name))
        self.modules.append(module)

    def add_modules(self, *args):
        for mod in args:
            self.add_module(mod)

    def add_arguments(self):
        subp = self.parser.add_subparsers()
        for mod in self.modules:
            modp = subp.add_parser(mod.name)
            mod.add_arguments(modp)
            modp.set_defaults(handler=mod.entry)

    def get_module(self, name):
        for mod in self.modules:
            if mod.name == name:
                return mod
        raise KeyError('no module named {}'.format(name))

    def get_selected(self, module):
        return self.get_module(module).get_selected()

    def iter_parent_modules(self, name):
        for mod in self.modules:
            if name in mod.dependencies:
                yield mod

    def iter_dependent_modules(self, parent):
        for mod in self.modules:
            if mod.name in parent.dependencies:
                yield mod

    def entry(self):
        self.add_arguments()
        self.args = self.parser.parse_args()
        try:
            handler = self.args.handler
        except AttributeError:
            handler = None
        if handler:
            handler(self.args)
        save_db(self.db)



import re
import json

from .module import Module


class Machine(Module):
    dependencies = ['app']
    ami_map = {
        # 'ap-southeast-2': 'ami-862211e5',  # docker enabled amazon
        'ap-southeast-2': 'ami-5c8fb13f',  # arch linux
    }

    def __init__(self, **kwargs):
        super().__init__('machine', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='machine help')

        p = subp.add_parser('list', help='list machines')
        p.add_argument('name', nargs='?')
        p.set_defaults(machine_handler=self.list)

        addp = subp.add_parser('add', help='add a machine')
        addp.add_argument('name', help='machine name')
        addp.add_argument('--public', '-p', action='store_true',
                          help='assign public IP')
        addp.set_defaults(machine_handler=self.add)

        p = subp.add_parser('ip', help='allocate a public IP')
        p.set_defaults(machine_handler=self.handle_ip)

        addp = subp.add_parser('init')
        addp.add_argument('name', help='machine name')
        addp.set_defaults(machine_handler=self.handle_init_swarm)

        sshp = subp.add_parser('ssh', help='ssh to a machine')
        sshp.add_argument('name', nargs='?', help='machine name')
        sshp.set_defaults(machine_handler=self.ssh)

        p = subp.add_parser('scp', help='scp')
        p.add_argument('src')
        p.add_argument('dst')
        p.set_defaults(machine_handler=self.scp)

        remp = subp.add_parser('remove', help='remove a machine')
        remp.add_argument('name', help='machine name')
        remp.set_defaults(machine_handler=self.remove)

        selp = subp.add_parser('select', help='select a machine')
        selp.add_argument('name', nargs='?', help='machine name')
        selp.add_argument('--show', '-s', action='store_true', help='show currently selected')
        selp.set_defaults(machine_handler=self.select)

        p = subp.add_parser('stats', help='show streaming stats')
        p.add_argument('name', nargs='?', help='machine name')
        p.set_defaults(machine_handler=self.handle_stats)

        p = subp.add_parser('reboot', help='reboot a machine')
        p.add_argument('name', nargs='?', help='machine name')
        p.set_defaults(machine_handler=self.handle_reboot)

    def handle_reboot(self, args):
        app = self.client.get_selected('app')
        name = args.name
        if name and not self.exists(name):
            self.error('unknown machine')
        if not name:
            name = self.get_selected()
        inst = self.get_instance(name, app)
        self.run(
            '$aws ec2 reboot-instances'
            ' --instance-ids {}'.format(
                inst['InstanceId']
            )
        )

    def tag_instance(self, inst_id, tags):
        tags = ['Key=%s,Value=%s' % (k, v) for k, v in tags.items()]
        self.run(
            '$aws ec2 create-tags'
            ' --resources {}'
            ' --tags {}'.format(
                inst_id,
                ' '.join(tags)
            )
        )

    def allocate_address(self, inst_id):
        data = self.run(
            '$aws ec2 allocate-address'
            ' --domain vpc',
            capture='json'
        )
        alloc_id = data['AllocationId']
        public_ip = data['PublicIp']
        self.run(
            '$aws ec2 associate-address'
            ' --instance-id {}'
            ' --allocation-id {}'.format(
                inst_id,
                alloc_id
            )
        )
        return alloc_id, public_ip

    def release_address(self, alloc_id):
        self.run(
            '$aws ec2 release-address'
            ' --allocation-id {}'.format(
                alloc_id
            )
        )

    def run_instance(self, name, region):
        image = self.ami_map[region]
        ctx = self.merged_config({'machine': name}, use_self=False)
        with self.template_file('arch-user-data.sh', ctx) as user_data:
            inst_id = self.run(
                '$aws ec2 run-instances'
                ' --image-id {image}'
                ' --key-name fuku-$app'
                ' --security-group-ids "$security_group"'
                ' --user-data file://{user_data}'
                ' --instance-type t2.micro'
                ' --iam-instance-profile Name=ec2-profile'
                ' --associate-public-ip-address'
                ' --count 1'
                ' --query \'Instances[0].InstanceId\''
                .format(
                    user_data=user_data,
                    image=image
                ),
                capture='json'
            )
        return inst_id

    def delete_instance(self, inst_id):
        self.run(
            '$aws ec2 terminate-instances'
            ' --instance-ids "{}"'.format(
                inst_id
            )
        )

    def list_instances(self, app):
        # TODO: Don't select machines that are terminating.
        data = self.run(
            '$aws ec2 describe-instances'
            ' --filters Name=tag:app,Values=$app Name=instance-state-name,Values=pending,running,shutting-down,stopping,stopped'
            ' --query \'Reservations[*].Instances[*].{id:InstanceId,tags:Tags[*]}\'',
            capture='json'
        )
        insts = []
        if data:
            for r in data:
                for i in r:
                    insts.append({
                        'id': i['id'],
                        'tags': dict([(t['Key'], t['Value']) for t in i['tags']])
                    })
        return insts

    def get_instance(self, name, app):
        data = self.run(
            '$aws ec2 describe-instances'
            ' --filters Name=tag:app,Values=$app Name=tag:name,Values=$machine',
            {'machine': name},
            capture='json'
        )
        if not data:
            self.error('unable to retrieve instance')
        return data['Reservations'][0]['Instances'][0]

    def get_address(self, inst_id):
        data = self.run(
            '$aws ec2 describe-addresses'
            ' --filters Name=instance-id,Values=$inst_id',
            {'inst_id': inst_id},
            capture='json'
        )
        if not data:
            self.error('unable to retrieve address')
        return data['Addresses'][0]

    def wait(self, inst_id):
        self.run(
            '$aws ec2 wait instance-status-ok'
            ' --instance-ids "{}"'.format(
                inst_id
            )
        )

    def list(self, args):
        app = self.client.get_selected('app')
        if args.name:
            inst = self.get_instance(args.name, app)
            print(json.dumps(inst, indent=2))
        else:
            machs = self.list_instances(app)
            for m in machs:
                print('{}:{}'.format(m['tags']['name'], m['id']))

    def add(self, args):
        if not self.db.get('profile', {}).get('bucket', None):
            self.error('must set profile bucket')
        region = self.client.get_selected('region')
        app = self.client.get_selected('app')
        name = args.name
        if name and self.exists(name):
            self.error('name already used')
        inst_id = self.run_instance(name, region)
        self.tag_instance(inst_id, {
            'Name': '%s-%s' % (app, name),
            'name': name,
            'app': app,
            'manager': 'true'
        })
        self.wait(inst_id)
        self.run(
            '$aws s3 cp '
            ' {}'
            ' s3://$bucket/fuku/$app/machines/{}.json'.format(
                self.data_path('machine.json'),
                name
            )
        )
        if args.public:
            alloc_id, public_ip = self.allocate_address(inst_id)
        self.init_swarm(args.name)

    def handle_init_swarm(self, args):
        self.init_swarm(args.name)

    def init_swarm(self, name):
        app = self.client.get_selected('app')
        inst = self.get_instance(name, app)
        ip = inst['PrivateIpAddress']
        response = self.ssh_run(
            'docker swarm init --advertise-addr {}'.format(ip),
            name=name,
            capture='text'
        )
        try:
            token = re.search(r'token\s+(.*)\s+\\', response).group(1)
            port = re.search(r':(\d\d\d\d)', response).group(1)
        except AttributeError:
            self.error('init failed')
        self.tag_instance(inst['InstanceId'], {
            'swarmtoken': token,
            'swarmport': port
        })
        self.ssh_run(
            'docker network create --driver overlay all',
            name=name,
            capture='discard'
        )

    def remove(self, args):
        # TODO: Give an option to back out. Maybe even two.
        if not self.db.get('profile', {}).get('bucket', None):
            self.error('must set profile bucket')
        app = self.client.get_selected('app')
        name = args.name
        if not self.exists(name):
            self.error('no such machine')
        inst = self.get_instance(name, app)
        addr = self.get_address(inst['InstanceId'])
        if addr:
            self.release_address(addr['AllocationId'])
        self.delete_instance(inst['InstanceId'])
        self.run(
            '$aws s3 rm '
            ' s3://$bucket/fuku/$app/machines/$name.json',
            {'name': name}
        )
        sel = self.get_selected(fail=False)
        if sel == name:
            del self.store['selected']
        try:
            del self.store.get('machines', {})[name]
        except KeyError:
            pass

    def select(self, args):
        if args.show:
            sel = self.get_selected(fail=False)
            if sel:
                print(sel)
        else:
            name = args.name
            if name:
                if not self.exists(name):
                    self.error('unkown machine')
                data = self.store.setdefault('machines', {}).setdefault(name, {})
                self.store['selected'] = name
            else:
                try:
                    del self.store['selected']
                except KeyError:
                    pass
            self.clear_parent_selections()

    def handle_stats(self, args):
        self.ssh_run('docker stats', args.name)

    def handle_ip(self, args):
        app = self.client.get_selected('app')
        name = self.get_selected()
        inst = self.get_instance(name, app)
        id = inst['InstanceId']
        self.allocate_address(id)

    def ssh(self, args):
        self.ssh_run('', args.name)

    def ssh_run(self, cmd, name=None, tty=False, capture=None):
        app = self.client.get_selected('app')
        name = name or self.get_selected()
        inst = self.get_instance(name, app)
        ip = inst['PublicIpAddress']
        full_cmd = 'ssh%s -o "StrictHostKeyChecking no" -i "$pem" root@%s %s' % (' -t' if tty else '', ip, cmd)
        return self.run(full_cmd, capture=capture)

    def scp(self, args):
        app = self.client.get_selected('app')
        name = self.get_selected()
        inst = self.get_instance(name, app)
        ip = inst['PublicIpAddress']
        self.run(
            'scp -o "StrictHostKeyChecking no" -i "$pem" {} root@{}:{}'.format(
                args.src,
                ip,
                args.dst
            ),
            capture=False
        )

    def exists(self, name):
        app = self.client.get_selected('app')
        machs = self.list_instances(app)
        for m in machs:
            if m['tags']['name'] == name:
                return True
        return False

    def get_selected(self, fail=True):
        sel = self.store.get('selected', None)
        if not sel and fail:
            self.error('no machine selected')
        return sel



import os
import json
# from tinydb import TinyDB


def get_rc_path():
    return os.path.expanduser('~/.fukurc')


def get_default_db():
    path = os.path.expanduser(os.path.join(get_rc_path(), 'db.json'))
    dir = os.path.dirname(path)
    try:
        os.makedirs(dir)
    except OSError:
        pass
    if os.path.exists(path):
        with open(path, 'r') as inf:
            db = json.load(inf)
    else:
        db = {}
    return db


def save_db(db):
    path = os.path.expanduser(os.path.join(get_rc_path(), 'db.json'))
    with open(path, 'w') as outf:
        json.dump(db, outf, indent=2)



import os
import stat

from .module import Module
from .db import get_rc_path
from .utils import entity_already_exists


class App(Module):
    dependencies = ['cluster']

    def __init__(self, **kwargs):
        super().__init__('app', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='app help')

        p = subp.add_parser('ls', help='list applications')
        p.set_defaults(app_handler=self.handle_list)

        p = subp.add_parser('mk', help='add an app')
        p.add_argument('name', metavar='NAME', help='app name')
        p.set_defaults(app_handler=self.handle_make)

        # remp = subp.add_parser('remove', help='remove an app')
        # remp.add_argument('name', help='app name')
        # remp.set_defaults(app_handler=self.remove)

        p = subp.add_parser('sl', help='select an app')
        p.add_argument('name', metavar='NAME', help='app name')
        p.set_defaults(app_handler=self.handle_select)

        p = subp.add_parser('run', help='run a command')
        p.add_argument('image', metavar='IMAGE', help='image name')
        p.add_argument('command', metavar='CMD', nargs='+', help='command to run')
        p.set_defaults(app_handler=self.handle_run)

    def handle_list(self, args):
        self.list()

    def list(self):
        for gr in self.iter_groups():
            print(gr.group_name[5:])

    def handle_make(self, args):
        self.make(args.name)

    def make(self, name):
        self.use_context = False
        self.create_group(name)
        self.make_task(name)
        self.select(name)

    def handle_select(self, args):
        self.select(args.name)

    def select(self, name):
        if name and name not in [g.group_name[5:] for g in self.iter_groups()]:
            self.error(f'no app "{name}"')
        self.store_set('selected', name)
        self.clear_parent_selections()

    def handle_run(self, args):
        self.run(args.image, args.command)

    def run(self, img, cmd):
        img = self.client.get_module('image').get_uri(img)
        cmd = ' '.join(cmd or [])
        full_cmd = f'docker run --rm -it {img} {cmd}'
        node_mod = self.client.get_module('node')
        node_mod.mgr_run(full_cmd, tty=True)

    def iter_groups(self):
        iam = self.get_boto_resource('iam')
        for gr in iam.groups.filter(PathPrefix='/fuku/'):
            yield gr

    def create_group(self, name):
        ctx = self.get_context()
        iam = self.get_boto_client('iam')
        with entity_already_exists():
            iam.create_group(
                Path=f'/fuku/{ctx["cluster"]}/{name}/',
                GroupName=f'fuku-{name}'
            )

    def delete_app_group(self, name):
        self.run(
            '$aws iam delete-group'
            ' --group-name fuku-$app',
            {'app': name}
        )

    def make_task(self, name):
        ctx = self.get_context()
        ecr_cli = self.get_boto_client('ecr')
        with entity_already_exists():
            ecr_cli.create_repository(
                repositoryName='fuku'
            )
        img_uri = self.client.get_module('image').image_name_to_uri('/fuku')
        task = {
            'family': f'fuku-{ctx["cluster"]}-{name}',
            'containerDefinitions': []
        }
        ctr_def = {
            'name': name,
            'image': img_uri,
            'memoryReservation': 1
        }
        task['containerDefinitions'].append(ctr_def)
        ecs = self.get_boto_client('ecs')
        skip = set(['taskDefinitionArn', 'revision', 'status', 'requiresAttributes'])
        ecs.register_task_definition(**dict([
            (k, v) for k, v in task.items() if k not in skip
        ]))

    def get_my_context(self):
        sel = self.store_get('selected')
        if not sel:
            import pdb; pdb.set_trace()
            self.error('no app currently selected')
        return {
            'app': sel
        }


class EcsApp(App):
    def make(self, name):
        super().make(name)
        self.make_target_group(name)

    def make_target_group(self, name):
        ctx = self.get_context()
        vpc_id = self.get_module('cluster').get_vpc(ctx['cluster']).id
        alb_cli = self.get_boto_client('elbv2')
        tg_arn = alb_cli.create_target_group(
            Name=f'fuku-{ctx["cluster"]}-{name}',
            Protocol='HTTP',
            Port=80,
            VpcId=vpc_id
        )['TargetGroups'][0]['TargetGroupArn']

    def get_target_group_arn(self):
        ctx = self.get_context()
        alb_cli = self.get_boto_client('elbv2')
        tg_arn = alb_cli.describe_target_groups(
            Names=[f'fuku-{ctx["cluster"]}-{ctx["app"]}']
        )['TargetGroups'][0]['TargetGroupArn']
        return tg_arn



import os

from .module import Module


class Region(Module):
    dependencies = ['profile']
    regions = set([
        'ap-southeast-2'
    ])

    def __init__(self, **kwargs):
        super().__init__('region', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='region help')

        p = subp.add_parser('ls', help='list regions')
        p.set_defaults(region_handler=self.handle_list)

        p = subp.add_parser('sl', help='select a region')
        p.add_argument('name', metavar='NAME', help='region name')
        p.set_defaults(region_handler=self.handle_select)

    def handle_list(self, args):
        self.list()

    def list(self):
        for r in sorted(list(self.regions)):
            print(r)

    def handle_select(self, args):
        self.select(args.name)

    def select(self, name):
        if name and name not in self.regions:
            self.error(f'no region "{name}"')
        self.store_set('selected', name)
        self.clear_parent_selections()

    def get_my_context(self):
        ctx = {}
        sel = self.get_selected()
        if sel:
            ctx['region'] = sel
        return ctx

    def get_selected(self):
        sel = self.store_get('selected')
        if not sel:
            self.error('no region currently selected')
        return sel
